import unittest

from crimson.templator import (
    format_insert,
    format_indent,
)


class TestSafeGuard(unittest.TestCase):
    def test_no_safe_flag_insert(self):
        kwargs = {"arg1": None, "arg2": "It might be an integer."}
        template = r"What is arg1? Arg1 is \[arg1\]. It \[arg2\]"

        with self.assertRaises(Exception):
            format_insert(template, **kwargs, safe=False)

    def test_no_safe_flag_indent(self):
        kwargs = {"arg1": "I am in line1", "arg2": None}
        template = r"""
            \{arg1\}

            \{arg2\}
        """

        with self.assertRaises(Exception):
            format_indent(template, **kwargs, safe=False)

    def test_safe_flag_insert(self):
        kwargs = {"arg1": None, "arg2": "might be an integer."}
        template = r"What is arg1? Arg1 is \[arg1\]. It \[arg2\]"
        expected_formatted = "What is arg1? Arg1 is None. It might be an integer."

        formatted = format_insert(template, **kwargs, safe=True)

        self.assertEqual(expected_formatted, formatted)

    def test_safe_flag_indent(self):
        kwargs = {"arg1": "I am in line1", "arg2": None}
        template = r"""
    \{arg1\}

    \{arg2\}
"""
        expected_formatted = """
    I am in line1

    None
"""

        formatted = format_indent(template, **kwargs, safe=True)
        self.assertEqual(expected_formatted, formatted)


if __name__ == "__main__":
    unittest.main()



from typing import Dict, Any, List, Union
from .utils import convert_lines, add_prefix, convert_dict_of_lists_to_list_of_dicts


def format_insert(
    # Text to convert.
    text: str,
    # Custom open bracket.
    open: str = r"\[",
    # Custom close bracket.
    close: str = r"\]",
    # Wrap values with str(value).
    # Otherwise, values such as 1, None, True and non-string values will cause an error.
    safe: bool = True,
    # Target name & new text pairs.
    # Pass it like
    # 'format_insert(..., kwarg1='value1', kwarg2='value2')' or
    # 'format_insert(..., **dictionary)'
    **kwargs: Dict[str, str],
):
    """
    String format function with custom brackets.

    Parameters
    ----------
    text: str,
        Text to convert.
    open: str = r"\[",
        Custom open bracket.
    close: str = r"\]",
        Custom close bracket.
    **kwargs
        Target name & new text pairs.
        Pass it like
        'format_insert(..., kwarg1='value1', kwarg2='value2')' or
        'format_insert(..., **dictionary)'
    """
    for key, value in kwargs.items():
        if safe:
            value = _convert_to_str(value)

        pattern = open + key + close
        text = text.replace(pattern, value)

    return text


def format_indent(
    # Text to convert.
    text: str,
    # Custom open bracket.
    open: str = r"\{",
    # Custom close bracket.
    close: str = r"\}",
    # Wrap values with str(value).
    # Otherwise, values such as 1, None, True and non-string values will cause an error.
    safe: bool = True,
    # Target name & new text pairs.
    # Pass it like
    # 'format_insert(..., kwarg1='value1', kwarg2='value2')' or
    # 'format_insert(..., **dictionary)'
    **kwargs: Dict[str, str],
):
    """
    String format function with custom brackets.
    The line with the target pattern can't have other letter than the pattern.

    Parameters
    ----------
    text: str,
        Text to convert.
    open: str = r"\{",
        Custom open bracket.
    close: str = r"\}",
        Custom close bracket.
    **kwargs,
        Target name & new text pairs.
        Pass it like
        'format_insert(..., kwarg1='value1', kwarg2='value2')' or
        'format_insert(..., **dictionary)'
    """
    for key, value in kwargs.items():
        if safe:
            value = _convert_to_str(value)
        text = _format_indent_single(text, key, value, open, close)

    return text


def format_insert_loop(
    template: str,
    kwargs: Union[Dict[str, List[str]], Dict[str, List[str]]],
    open: str = r"\\[",
    close: str = r"\\]",
    safe: bool = True,
):
    parsers = [format_insert_loop_many, format_insert_loop_list]
    errors = []

    for parser in parsers:
        try:
            return parser(template, kwargs, open, close, safe)
        except Exception as e:
            errors.append(f"{parser.__name__} error: {e}")
            continue

    raise ValueError(
        "Both format_insert_loop_many and format_insert_loop_legacy failed with errors: " + "; ".join(errors)
    )


def format_insert_loop_many(
    template: str,
    kwargs_many: Dict[str, List[str]],
    open: str = r"\\[",
    close: str = r"\\]",
    safe: bool = True,
):
    kwargs_list = convert_dict_of_lists_to_list_of_dicts(kwargs_many)

    return format_insert_loop_list(
        template,
        kwargs_list,
        open,
        close,
        safe,
    )


def format_insert_loop_list(
    template: str,
    kwargs_list: List[Dict[str, str]],
    open: str = r"\\[",
    close: str = r"\\]",
    safe: bool = True,
):
    formatted_lines = []

    for kwargs in kwargs_list:
        formatted = format_insert(template, open, close, safe, **kwargs)
        formatted_lines.append(formatted)
    formatted = convert_lines(formatted_lines)
    return formatted


def _convert_to_str(value: Any):
    return str(value)


def _format_indent_single(
    text: str,
    key: str,
    value: str,
    open: str = r"\{",
    close: str = r"\}",
):
    pattern = open + key + close
    new_lines = []
    for line in text.split("\n"):
        if line.find(pattern) != -1:
            _check_indent_line(line, pattern)
            indent = line[: line.find(pattern)]
            new_lines.append(add_prefix(value, indent))
        else:
            new_lines.append(line)

    return "\n".join(new_lines)


def _check_indent_line(text: str, pattern: str):
    remaining_text = text.replace(pattern, "").strip()

    if remaining_text:
        raise ValueError(f"The line contains characters other than '{pattern}'")

    return True



from typing import List, Any, Iterable, Dict


def convert_lines(lines: List[str]) -> str:
    return "\n".join(lines)


def convert_list_to_dicts_list(inputs: Iterable[Any], shared_key: str) -> List[Dict[str, str]]:
    kwargs = [{shared_key: input} for input in inputs]
    return kwargs


def add_prefix(
    text: str,
    prefix: str = " " * 4,
):
    """
    Parameters
    ----------
    text: str,
        Text to convert
    prefix: str = " " * 4
        Prefix to be added to all the lines
    """
    split = text.split("\n")
    text = prefix + f"\n{prefix}".join(split)
    return text


def convert_dict_of_lists_to_list_of_dicts(
    dict_of_lists: Dict[str, List[str]],
) -> List[Dict[str, str]]:
    if not _is_dict_of_lists_valid(dict_of_lists):
        raise ValueError("All lists in dict_of_lists must have the same length")

    count = len(next(iter(dict_of_lists.values())))
    list_of_dicts = [{key: value_list[i] for key, value_list in dict_of_lists.items()} for i in range(count)]
    return list_of_dicts


def convert_list_of_dicts_to_dict_of_lists(
    list_of_dicts: List[Dict[str, str]],
) -> Dict[str, List[str]]:
    if not list_of_dicts:
        return {}

    keys = list_of_dicts[0].keys()
    dict_of_lists = {key: [] for key in keys}

    for d in list_of_dicts:
        for key in keys:
            dict_of_lists[key].append(d[key])

    return dict_of_lists


def _is_dict_of_lists_valid(
    kwargs_many: Dict[str, List[str]],
) -> bool:
    lengths = iter(len(args) for args in kwargs_many.values())
    first_length = next(lengths, None)
    return all(length == first_length for length in lengths)



import unittest
from pydantic import BaseModel
from typing import List, Dict

from crimson.templator import format_insert_loop_many, format_insert_loop_list
from crimson.templator.utils import (
    convert_list_to_dicts_list,
)


class TestFormatInsertLoop(unittest.TestCase):

    kwargs1 = {
        "name": "Amy",
        "age": "25",
        "address": "Erlangen",
    }

    kwargs2 = {
        "name": "Jone",
        "age": "13",
        "address": "London",
    }

    kwargs_list = [kwargs1, kwargs2]

    kwargs_many = {"name": ["Amy", "Jone"], "age": ["25", "13"], "address": ["Erlangen", "London"]}

    template = r"""{
    name : \\[name\\],
    age : \\[age\\],
    address : \\[address\\],
},"""
    expected_formatted = """{
    name : Amy,
    age : 25,
    address : Erlangen,
},
{
    name : Jone,
    age : 13,
    address : London,
},"""

    def test_loop_kwargs_list(self):

        formatted = format_insert_loop_many(template=self.template, kwargs_many=self.kwargs_many)

        self.assertEqual(formatted, self.expected_formatted)

    def test_format_loop_list(self):
        # Action
        formatted = format_insert_loop_list(template=self.template, kwargs_list=self.kwargs_list)
        print(formatted)

        # Assertion
        self.assertEqual(formatted, self.expected_formatted)


if __name__ == "__main__":
    unittest.main()



import unittest

from crimson.templator import (
    format_indent,
)


class TestSafeGuard(unittest.TestCase):
    def test_no_safe_flag_indent(self):
        kwargs = {
            "arg1": """\
I want to write very long lines
1
2
3
even 4!\
"""
        }
        template = r"""
    \{arg1\}
"""
        expected_formatted = """
    I want to write very long lines
    1
    2
    3
    even 4!
"""

        formatted = format_indent(template, **kwargs)

        self.assertAlmostEqual(expected_formatted, formatted)

    def test_one_line_only_one_indent(self):
        kwargs = {"arg1": "I am just a line."}
        template = r"""
    \{arg1\} Additional text with indent will cause an error.
"""

        with self.assertRaises(Exception):
            format_indent(template, **kwargs, safe=True)


if __name__ == "__main__":
    unittest.main()
